from dataclasses import dataclass
from decimal import Decimal
from typing import Any, Callable, Dict, Optional, Type, Union

from koda_validate import NotBlank
from koda_validate.base import Predicate, PredicateAsync
from koda_validate.dataclasses import DataclassValidator
from koda_validate.decimal import DecimalValidator
from koda_validate.dictionary import MaxKeys, MinKeys
from koda_validate.errors import (
    CoercionErr,
    ContainerErr,
    ExtraKeysErr,
    IndexErrs,
    KeyErrs,
    MapErr,
    MissingKeyErr,
    PredicateErrs,
    SetErrs,
    TypeErr,
    UnionErrs,
    ValidationErrBase,
)
from koda_validate.generic import (
    Choices,
    EndsWith,
    EqualTo,
    ExactItemCount,
    ExactLength,
    Max,
    MaxItems,
    MaxLength,
    Min,
    MinItems,
    MinLength,
    MultipleOf,
    StartsWith,
    UniqueItems,
)
from koda_validate.namedtuple import NamedTupleValidator
from koda_validate.serialization.base import Serializable
from koda_validate.string import EmailPredicate, RegexPredicate
from koda_validate.time import DatetimeValidator, DateValidator
from koda_validate.uuid import UUIDValidator
from koda_validate.valid import Invalid


@dataclass
class SerializableErr(ValidationErrBase):
    """
    A validation error where you may simply want to produce a human-readable message.
    """

    obj: Serializable
    """
    The error contents you want to return.
    """


def pred_to_err_message(pred: Union[Predicate[Any], PredicateAsync[Any]]) -> str:
    if isinstance(pred, MinKeys):
        return f"minimum allowed properties is {pred.size}"
    elif isinstance(pred, MaxKeys):
        return f"maximum allowed properties is {pred.size}"
    elif isinstance(pred, Choices):
        return f"expected one of {sorted(pred.choices)}"
    elif isinstance(pred, Min):
        exclusive = " (exclusive)" if pred.exclusive_minimum else ""
        return f"minimum allowed value{exclusive} is {pred.minimum}"
    elif isinstance(pred, Max):
        exclusive = " (exclusive)" if pred.exclusive_maximum else ""
        return f"maximum allowed value{exclusive} is {pred.maximum}"
    elif isinstance(pred, MultipleOf):
        return f"expected multiple of {pred.factor}"
    elif isinstance(pred, EqualTo):
        return f"must equal {repr(pred.match)}"
    elif isinstance(pred, MinItems):
        return f"minimum allowed length is {pred.item_count}"
    elif isinstance(pred, MaxItems):
        return f"maximum allowed length is {pred.item_count}"
    elif isinstance(pred, ExactItemCount):
        return f"length must be {pred.item_count}"
    elif isinstance(pred, UniqueItems):
        return "all items must be unique"
    elif isinstance(pred, RegexPredicate):
        return rf"must match pattern {pred.pattern.pattern}"
    elif isinstance(pred, EmailPredicate):
        return "expected a valid email address"
    elif isinstance(pred, NotBlank):
        return "cannot be blank"
    elif isinstance(pred, MinLength):
        return f"minimum allowed length is {pred.length}"
    elif isinstance(pred, MaxLength):
        return f"maximum allowed length is {pred.length}"
    elif isinstance(pred, ExactLength):
        return f"expected length of {pred.length}"
    elif isinstance(pred, StartsWith):
        return rf"must start with {repr(pred.prefix)}"
    elif isinstance(pred, EndsWith):
        return rf"must end with {repr(pred.suffix)}"
    else:
        raise TypeError(
            f"Unhandled predicate type: {type(pred)}. You may want to write a wrapper "
            f"function which handles that type."
        )


TYPE_DESCRIPTION_LOOKUP: Dict[Type[Any], str] = {
    str: "string",
    int: "integer",
    Decimal: "decimal",
    bool: "boolean",
    dict: "object",
    list: "array",
    tuple: "array",
}


def to_serializable_errs(
    invalid: Invalid, next_level: Optional[Callable[[Invalid], Serializable]] = None
) -> Serializable:
    """
    Convert an ``Invalid`` instance to human readable (English) ``Serializable``
    representation of ``Invalid``.

    It can serve as an example of how to build similar functions to convert ``Invalid``
    instances to other formats, other languages, etc.

    :param invalid: The error you'd like to represent
    :param next_level: If supplied, this callable will handle any calls for container
        ``ErrType``s such as ``ContainerErr``, ``KeyErr``, and so on

    :return: an error message appropriate for serializing in JSON or YAML. Because this
        function can handle any kind of Invalid, any type narrowing desired beyond
        ``Serializable`` needs to be done outside of this function.

    """
    next_level = next_level or to_serializable_errs
    err = invalid.err_type
    vldtr = invalid.validator
    if isinstance(err, CoercionErr):
        if isinstance(vldtr, UUIDValidator):
            return ["expected a UUID"]
        elif isinstance(vldtr, DecimalValidator):
            return ["expected a decimal-formatted string"]
        elif isinstance(vldtr, DatetimeValidator):
            return ["expected an iso8601 datetime string"]
        elif isinstance(vldtr, DateValidator):
            return ["expected YYYY-MM-DD"]
        elif err.dest_type is list or err.dest_type is tuple:
            return {"__container__": ["expected a list"]}
        elif isinstance(vldtr, (DataclassValidator, NamedTupleValidator)):
            return {"__container__": ["expected a dict"]}
        else:
            compatible_names = sorted([t.__name__ for t in err.compatible_types])
            return [
                f"could not coerce to {err.dest_type.__name__} "
                f"(compatible with {', '.join(compatible_names)})"
            ]
    elif isinstance(err, SerializableErr):
        return err.obj
    elif isinstance(err, ExtraKeysErr):
        err_message = (
            "expected an empty dict"
            if len(err.expected_keys) == 0
            else "only expected "
            + ", ".join(sorted([repr(k) for k in err.expected_keys]))
        )
        return {"__unknown_keys__": err_message}
    elif isinstance(err, TypeErr):
        if err.expected_type is dict:
            return {"__container__": ["expected a dict"]}
        elif err.expected_type is list or err.expected_type is tuple:
            return {"__container__": ["expected a list"]}
        else:
            type_desc = TYPE_DESCRIPTION_LOOKUP.get(
                err.expected_type, err.expected_type.__name__
            )
            if type_desc[0] in {"a", "e", "i", "o", "u"}:
                type_desc = f"an {type_desc}"
            else:
                type_desc = f"a {type_desc}"
            return [f"expected {type_desc}"]
    elif isinstance(err, PredicateErrs):
        return [pred_to_err_message(p) for p in err.predicates]
    elif isinstance(err, IndexErrs):
        return [[i, next_level(err)] for i, err in err.indexes.items()]
    elif isinstance(err, MissingKeyErr):
        return ["key missing"]
    elif isinstance(err, MapErr):
        errs_dict: Dict[str, Serializable] = {}
        for key, k_v_errs in err.keys.items():
            kv_dict: Dict[str, Serializable] = {
                k: next_level(v)
                for k, v in [("key", k_v_errs.key), ("value", k_v_errs.val)]
                if v is not None
            }
            errs_dict[str(key)] = kv_dict
        return errs_dict
    elif isinstance(err, SetErrs):
        return {"member_errors": [next_level(x) for x in err.item_errs]}
    elif isinstance(err, KeyErrs):
        return {str(k): next_level(v) for k, v in err.keys.items()}
    elif isinstance(err, UnionErrs):
        return {"variants": [next_level(x) for x in err.variants]}
    elif isinstance(err, ContainerErr):
        return next_level(err.child)
    else:
        raise TypeError(f"got unhandled type: {type(err)}")



import inspect
import sys
from dataclasses import is_dataclass
from datetime import date, datetime
from decimal import Decimal

from koda import Just, Nothing

from .is_type import TypeValidator
from .maybe import MaybeValidator

if sys.version_info >= (3, 9):
    from typing import Annotated

if sys.version_info >= (3, 10):
    from types import UnionType

if sys.version_info >= (3, 11):
    from typing import NotRequired, Required

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Set,
    Tuple,
    Union,
    get_args,
    get_origin,
)
from uuid import UUID

from ._internal import _is_typed_dict_cls
from .base import Validator
from .boolean import BoolValidator
from .bytes import BytesValidator
from .decimal import DecimalValidator
from .dictionary import MapValidator
from .float import FloatValidator
from .generic import Choices, EqualsValidator, always_valid
from .integer import IntValidator
from .list import ListValidator
from .none import NoneValidator, none_validator
from .set import SetValidator
from .string import StringValidator
from .time import DatetimeValidator, DateValidator
from .tuple import NTupleValidator, UniformTupleValidator
from .union import UnionValidator
from .uuid import UUIDValidator


def annotation_is_naked_tuple(annotation: Any) -> bool:
    return annotation is Tuple or annotation is tuple


def annotation_is_naked_list(annotation: Any) -> bool:
    return annotation is List or annotation is list


def annotation_is_namedtuple(annotation: Any) -> bool:
    return bool(
        (bases := getattr(annotation, "__bases__", None))
        and bases == (tuple,)
        and hasattr(annotation, "_fields")
    )


# todo: evolve into general-purpose type-hint driven validator
# will probably need significant changes
def get_typehint_validator_base(
    get_hint_next_depth: Callable[[Any], Validator[Any]], annotation: Any
) -> Validator[Any]:
    """
    get_hint_next_depth allows a developer to wrap this function but still have it
    use their wrapper recursively (otherwise we might simply recur to this function in
    each depth)

    :param get_hint_next_depth: the ``Callable`` that will determine the ``Validator``
        resolution at the next depth (usually this is the same at all depths)
    :param annotation: the type we're using to create a ``Validator``
    :return: a ``Validator`` that expresses the intent of the ``annotations``
    :raises TypeError: on types that are not handled
    """
    if annotation is str:
        return StringValidator()
    elif annotation is int:
        return IntValidator()
    elif annotation is float:
        return FloatValidator()
    elif annotation is None or annotation is type(None):  # noqa: E721
        return NoneValidator()
    elif annotation is UUID:
        return UUIDValidator()
    elif annotation is date:
        return DateValidator()
    elif annotation is datetime:
        return DatetimeValidator()
    elif annotation is bool:
        return BoolValidator()
    elif annotation is Decimal:
        return DecimalValidator()
    elif annotation is bytes:
        return BytesValidator()
    elif annotation is Any:
        return always_valid
    elif annotation_is_naked_list(annotation):
        return ListValidator(always_valid)
    elif annotation is Set or annotation is set:
        return SetValidator(always_valid)
    elif annotation_is_naked_tuple(annotation):
        return UniformTupleValidator(always_valid)
    elif annotation is Dict or annotation is dict:
        return MapValidator(key=always_valid, value=always_valid)
    elif is_dataclass(annotation):
        from .dataclasses import DataclassValidator

        return DataclassValidator(annotation)
    elif annotation_is_namedtuple(annotation):
        from .namedtuple import NamedTupleValidator

        return NamedTupleValidator(annotation)
    elif _is_typed_dict_cls(annotation):
        from .typeddict import TypedDictValidator

        return TypedDictValidator(annotation)
    else:
        origin, args = get_origin(annotation), get_args(annotation)
        if annotation_is_naked_list(origin) and len(args) == 1:
            item_validator = get_hint_next_depth(args[0])
            return ListValidator(item_validator)
        elif (origin is set or origin is Set) and len(args) == 1:
            item_validator = get_hint_next_depth(args[0])
            return SetValidator(item_validator)
        elif (origin is dict or origin is Dict) and len(args) == 2:
            return MapValidator(
                key=get_hint_next_depth(args[0]), value=get_hint_next_depth(args[1])
            )
        elif origin is Union or (sys.version_info >= (3, 10) and origin is UnionType):
            if len(args) == 2 and args[1] is Nothing and get_origin(args[0]) is Just:
                return MaybeValidator(get_hint_next_depth(get_args(args[0])[0]))
            else:
                return UnionValidator(*[get_hint_next_depth(arg) for arg in args])
        elif origin is tuple or origin is Tuple:
            if len(args) == 2 and args[1] is Ellipsis:
                return UniformTupleValidator(get_hint_next_depth(args[0]))
            else:
                return NTupleValidator.untyped(
                    fields=tuple(get_hint_next_depth(a) for a in args)
                )

        elif origin is Literal:
            # The common case is that literals will be of the same type. For those
            # cases, we return the relevant type validator (if we have it).
            if len(args) == 0:
                ValueError("Literal types cannot be empty")
            else:
                type_ = type(args[0])
                for a in args[1:]:
                    if type(a) is not type_:
                        break
                else:
                    if type_ is str:
                        return StringValidator(Choices(set(args)))
                    elif type_ is int:
                        return IntValidator(Choices(set(args)))
                    elif type_ is bool:
                        return BoolValidator(Choices(set(args)))
                    elif type_ is bytes:
                        return BytesValidator(Choices(set(args)))
                    elif type_ is type(None) or type_ is None:  # noqa: E721
                        return none_validator

            # ok, we have to use a union because we have multiple types
            # or we haven't defined an explicit validator to use
            return UnionValidator(*[EqualsValidator(a) for a in args])

        elif sys.version_info >= (3, 11) and (
            origin is NotRequired or origin is Required
        ):
            return get_typehint_validator(args[0])

        # not validating with annotations at this point
        elif sys.version_info >= (3, 9) and origin is Annotated:
            if len(args) == 1:
                return get_typehint_validator(args[0])
            else:
                # only return the first annotation validator
                for x in args:
                    if isinstance(x, Validator):
                        return x
        elif inspect.isclass(annotation):
            # fall back to an explicit type checker
            return TypeValidator(annotation)

        raise TypeError(f"Got unhandled annotation: {repr(annotation)}.")


def get_typehint_validator(annotation: Any) -> Validator[Any]:
    """
    The "default" way to convert typehints to `Validator`

    :param annotation: Any valid python annotation.
    :returns: a validator if it finds a match
    """
    return get_typehint_validator_base(get_typehint_validator, annotation)



import re
from dataclasses import dataclass
from typing import Pattern

from koda_validate._internal import _ToTupleStandardValidator
from koda_validate.base import Predicate


class StringValidator(_ToTupleStandardValidator[str]):
    r"""
    Validate a value is a ``str``, and any extra refinement.

    If ``predicates_async`` is supplied, the ``__call__`` method should not be
    called -- only ``.validate_async`` should be used.

    Example:

    >>> from koda_validate import *
    >>> validator = StringValidator(not_blank, MaxLength(100), preprocessors=[strip])
    >>> validator("")
    Invalid(err_type=PredicateErrs(predicates=[NotBlank()]), ...)
    >>> validator(None)
    Invalid(err_type=TypeErr(expected_type=<class 'str'>), ...)
    >>> validator(" ok ")
    Valid(val='ok')

    :param predicates: any number of ``Predicate[str]`` instances
    :param predicates_async: any number of ``PredicateAsync[str]`` instances
    :param preprocessors: any number of ``Processor[str]``, which will be run before
        :class:`Predicate`\s and :class:`PredicateAsync`\s are checked.
    """

    _TYPE = str


@dataclass
class RegexPredicate(Predicate[str]):
    pattern: Pattern[str]

    def __call__(self, val: str) -> bool:
        return self.pattern.match(val) is not None


@dataclass
class EmailPredicate(Predicate[str]):
    pattern: Pattern[str] = re.compile("[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+")

    def __call__(self, val: str) -> bool:
        return self.pattern.match(val) is not None



import decimal
from decimal import Decimal
from typing import Any, List, Optional

from koda import Just, Maybe, nothing

from koda_validate._internal import _ToTupleStandardValidator
from koda_validate.base import Predicate, PredicateAsync, Processor
from koda_validate.coerce import Coercer, coercer


@coercer(str, int, Decimal)
def coerce_decimal(val: Any) -> Maybe[Decimal]:
    if type(val) is Decimal:
        return Just(val)
    elif isinstance(val, (str, int)):
        try:
            return Just(Decimal(val))
        except decimal.InvalidOperation:
            pass

    return nothing


class DecimalValidator(_ToTupleStandardValidator[Decimal]):
    _TYPE = Decimal

    def __init__(
        self,
        *predicates: Predicate[Decimal],
        predicates_async: Optional[List[PredicateAsync[Decimal]]] = None,
        preprocessors: Optional[List[Processor[Decimal]]] = None,
        coerce: Optional[Coercer[Decimal]] = coerce_decimal,
    ) -> None:
        super().__init__(
            *predicates,
            predicates_async=predicates_async,
            preprocessors=preprocessors,
            coerce=coerce,
        )



from typing import (
    Any,
    Awaitable,
    Callable,
    List,
    Literal,
    NoReturn,
    Optional,
    Tuple,
    Type,
    Union,
)

from koda_validate._generics import A, SuccessT
from koda_validate.base import Predicate, PredicateAsync, Processor, Validator
from koda_validate.coerce import Coercer
from koda_validate.errors import CoercionErr, PredicateErrs, TypeErr, UnionErrs
from koda_validate.valid import Invalid, Valid, ValidationResult

_ResultTuple = Union[Tuple[Literal[True], A], Tuple[Literal[False], Invalid]]


class _ToTupleValidator(Validator[SuccessT]):
    """
    This `Validator` subclass exists for optimization. When we call
    nested validators it's much less computation to deal with simple
    tuples and bools, instead of Valid and Invalid instances.
    This class may go away!

    DO NOT USE THIS UNLESS YOU:
    - ARE OK WITH THIS DISAPPEARING IN A FUTURE RELEASE
    - ARE GOING TO TEST YOUR CODE EXTENSIVELY
    """

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[SuccessT]:
        raise NotImplementedError()  # pragma: no cover

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[SuccessT]:
        raise NotImplementedError()  # pragma: no cover

    async def validate_async(self, val: Any) -> ValidationResult[SuccessT]:
        result = await self._validate_to_tuple_async(val)
        if result[0]:
            return Valid(result[1])
        else:
            return result[1]

    def __call__(self, val: Any) -> ValidationResult[SuccessT]:
        result = self._validate_to_tuple(val)
        if result[0]:
            return Valid(result[1])
        else:
            return result[1]


def _simple_type_validator(
    instance: "_ToTupleStandardValidator[A]", type_: Type[A], type_err: TypeErr
) -> Callable[[Any], _ResultTuple[A]]:
    def inner(val: Any) -> _ResultTuple[A]:
        if type(val) is type_:
            return True, val
        else:
            return False, Invalid(type_err, val, instance)

    return inner


def _async_predicates_warning(cls: Type[Any]) -> NoReturn:
    raise AssertionError(
        f"{cls.__name__} cannot run `predicates_async` in synchronous calls. "
        f"Please `await` the `.validate_async` method instead; or remove the "
        f"items in `predicates_async`."
    )


def _raise_validate_object_async_in_sync_mode(cls: Type[Any]) -> NoReturn:
    raise AssertionError(
        f"{cls.__name__} cannot run `validate_object_async` in synchronous calls. "
        f"Please `await` the `.validate_async` method instead."
    )


def _repr_helper(cls: Type[Any], arg_strs: List[str]) -> str:
    return f"{cls.__name__}({', '.join(arg_strs)})"


class _ToTupleStandardValidator(_ToTupleValidator[SuccessT]):
    """
    This `Validator` subclass exists primarily for code cleanliness and standardization.
    It allows us to have very simple Scalar validators.

    This class may go away!

    DO NOT USE THIS UNLESS YOU:
    - ARE OK WITH THIS DISAPPEARING IN A FUTURE RELEASE
    - ARE GOING TO TEST YOUR CODE EXTENSIVELY
    """

    __match_args__ = ("predicates", "predicates_async", "preprocessors")

    # SHOULD BE THE SAME AS SuccessT but mypy can't handle that...? v0.991
    _TYPE: Type[Any]
    _type_err: TypeErr

    def __init__(
        self,
        *predicates: Predicate[SuccessT],
        predicates_async: Optional[List[PredicateAsync[SuccessT]]] = None,
        preprocessors: Optional[List[Processor[SuccessT]]] = None,
        coerce: Optional[Coercer[SuccessT]] = None,
    ) -> None:
        self.predicates = predicates
        self.predicates_async = predicates_async
        self.preprocessors = preprocessors
        _type_err = TypeErr(self._TYPE)
        self._type_err = _type_err
        self._disallow_synchronous = bool(predicates_async)
        self.coerce = coerce

        # optimization for simple  validators. can speed up by ~15%
        if not predicates and not predicates_async and not preprocessors and not coerce:
            self._validate_to_tuple = _simple_type_validator(  # type: ignore
                self, self._TYPE, _type_err
            )

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[SuccessT]:
        if self._disallow_synchronous:
            _async_predicates_warning(self.__class__)

        if self.coerce:
            result = self.coerce(val)
            if not result.is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, self._TYPE),
                    val,
                    self,
                )
            else:
                # val is now SuccessT
                val = result.val
        elif type(val) is not self._TYPE:
            return False, Invalid(self._type_err, val, self)

        if self.preprocessors:
            for proc in self.preprocessors:
                val = proc(val)

        if self.predicates:
            errors: List[Any] = [pred for pred in self.predicates if not pred(val)]
            if errors:
                return False, Invalid(PredicateErrs(errors), val, self)
            else:
                return True, val
        else:
            return True, val

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[SuccessT]:
        if self.coerce:
            result = self.coerce(val)
            if not result.is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, self._TYPE),
                    val,
                    self,
                )
            else:
                # val is now SuccessT
                val = result.val

        elif type(val) is not self._TYPE:
            return False, Invalid(self._type_err, val, self)

        if self.preprocessors:
            for proc in self.preprocessors:
                val = proc(val)

        errors: List[Union[Predicate[SuccessT], PredicateAsync[SuccessT]]] = [
            pred for pred in self.predicates if not pred(val)
        ]

        if self.predicates_async:
            errors.extend(
                [
                    pred
                    for pred in self.predicates_async
                    if not (await pred.validate_async(val))
                ]
            )
        if errors:
            return False, Invalid(PredicateErrs(errors), val, self)
        else:
            return True, val

    def __eq__(self, other: Any) -> bool:
        return (
            type(other) is type(self)
            and self.predicates == other.predicates
            and self.predicates_async == other.predicates_async
            and self.preprocessors == other.preprocessors
        )

    def __repr__(self) -> str:
        return _repr_helper(
            self.__class__,
            ([] if not self.predicates else [repr(p) for p in self.predicates])
            + [
                f"{k}={repr(v)}"
                for k, v in [
                    ("predicates_async", self.predicates_async),
                    ("preprocessors", self.preprocessors),
                ]
                if v
            ],
        )


def _union_validator(
    source_validator: Validator[A], validators: Tuple[Validator[Any], ...], val: Any
) -> _ResultTuple[A]:
    errs = []
    for validator in validators:
        if isinstance(validator, _ToTupleValidator):
            result_tup = validator._validate_to_tuple(val)
            if result_tup[0]:
                return True, result_tup[1]
            else:
                errs.append(result_tup[1])
        else:
            result = validator(val)
            if result.is_valid:
                return True, result.val
            else:
                errs.append(result)
    return False, Invalid(UnionErrs(errs), val, source_validator)


async def _union_validator_async(
    source_validator: Validator[A], validators: Tuple[Validator[Any], ...], val: Any
) -> _ResultTuple[A]:
    errs = []
    for validator in validators:
        if isinstance(validator, _ToTupleValidator):
            result_tup = await validator._validate_to_tuple_async(val)
            if result_tup[0]:
                return True, result_tup[1]
            else:
                errs.append(result_tup[1])
        else:
            result = await validator.validate_async(val)
            if result.is_valid:
                return True, result.val
            else:
                errs.append(result)
    return False, Invalid(UnionErrs(errs), val, source_validator)


def _wrap_sync_validator(obj: Validator[A]) -> Callable[[Any], _ResultTuple[A]]:
    if isinstance(obj, _ToTupleValidator):
        return obj._validate_to_tuple
    else:

        def inner(v: Any) -> _ResultTuple[A]:
            result = obj(v)
            if result.is_valid:
                return True, result.val
            else:
                return False, result

        return inner


def _wrap_async_validator(
    obj: Validator[A],
) -> Callable[[Any], Awaitable[_ResultTuple[A]]]:
    if isinstance(obj, _ToTupleValidator):
        return obj._validate_to_tuple_async
    else:
        async_validator = obj.validate_async

        async def inner(v: Any) -> _ResultTuple[A]:
            result = await async_validator(v)
            if result.is_valid:
                return True, result.val
            else:
                return False, result

        return inner


def _is_typed_dict_cls(t: Type[Any]) -> bool:
    return (
        hasattr(t, "__annotations__") and hasattr(t, "__total__") and hasattr(t, "keys")
    )


def _raise_cannot_define_validate_object_and_validate_object_async() -> None:
    raise AssertionError(
        "validate_object and validate_object_async cannot both be defined"
    )



from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic

from koda import Maybe

from koda_validate._generics import A, SuccessT

if TYPE_CHECKING:
    from koda_validate.valid import ValidationResult


class Validator(Generic[SuccessT]):
    r"""
    Base class for all ``Validator``\s.

    It's little more than a ``Callable[[Any], ValidationResult[SuccessT]]``, with
    two notable differences:

    - A ``.validate_async`` method is allowed, meaning a :class:`Validator` can be made
        to work in both sync and async contexts.
    - :class:`Validator`s are ``class``\es. Constructing ``Callable`` ``class``\es allows
        us to more easily make metadata from the validator available (as opposed to data
        being hidden inside a closure)

    Depending on your use case, you may want to implement the ``__call__`` method, the
    ``async_validate`` method, or both.

    .. note::

        Fundamentally, :class:`Validator`\s need to be able to return valid data that is
        a different type than what was passed in. The main reason for this is that we
        need to be able to call a ``Validator`` with data of unknown types. For that
        reason ``Validator``\s accept ``Any`` as input. This has a few notable
        implications:

        - we can never assume that the data returned is the same data passed in
        - coercing and conforming (see :class:`Processor`\s) data is fundamentally legal
        - some kind of type verification usually needs to happen within a ``Validator``
        - refinement (see :class:`Predicate`\s) needs to be run *after* type verification


    """

    async def validate_async(self, val: Any) -> "ValidationResult[SuccessT]":
        """
        :param val: the value being validated
        """
        raise NotImplementedError()  # pragma: no cover

    def __call__(self, val: Any) -> "ValidationResult[SuccessT]":
        """
        :param val: the value being validated
        """

        raise NotImplementedError()  # pragma: no cover


class Predicate(Generic[A]):
    r"""
    A predicate just returns ``True`` or ``False`` for some condition.

    :class:`Predicate`\s  can be used during async validation, but
    :class:`PredicateAsync` should be used for any validation that *requires*
    ``asyncio``.

    Example :class:`Predicate`:

    .. testcode:: predsubclass

        from koda_validate import Predicate

        class GreaterThan(Predicate[int]):
            def __init__(self, limit: int) -> None:
                self.limit = limit

            def __call__(self, val: int) -> bool:
                return val > self.limit

    Usage

    .. doctest:: predsubclass

        >>> gt = GreaterThan(5)
        >>> gt(6)
        True
        >>> gt(1)
        False
    """

    @abstractmethod
    def __call__(self, val: A) -> bool:  # pragma: no cover
        """
        :param val: the value being validated
        """
        raise NotImplementedError()  # pragma: no cover


class PredicateAsync(Generic[A]):
    """
    The async-only sibling of :class:`Predicate`.

    Example :class:`PredicateAsync`

    .. testcode:: predasyncsubclass

        import asyncio
        from koda_validate import PredicateAsync

        class UsernameInDB(PredicateAsync[str]):
            async def validate_async(self, val: str) -> bool:
                # pretend to call db
                await asyncio.sleep(.001)
                # dummy logic for example
                return len(val) == 3

    Usage

    .. doctest:: predasyncsubclass

        >>> asyncio.run(UsernameInDB().validate_async("abc"))
        True
        >>> asyncio.run(UsernameInDB().validate_async("abcdef"))
        False
    """

    @abstractmethod
    async def validate_async(self, val: A) -> bool:  # pragma: no cover
        """
        :param val: the value being validated
        :return: a bool indicating whether the condition is ``True`` for the value
        """
        raise NotImplementedError()  # pragma: no cover


class Processor(Generic[A]):
    r"""
    Base class for ``Processor``\s. These are litle more than

    .. code-block::

        Callable[[A], A]


    They transform a value of one type to another value of the same type. The
    are useful for things like ``strip``\-ping strings:

    .. testcode:: strip

        from dataclasses import dataclass
        from koda_validate import Processor

        @dataclass
        class Strip(Processor[str]):
            def __call__(self, val: str) -> str:
                return val.strip()


    Usage:

    .. doctest:: strip

        >>> strip = Strip()
        >>> strip(" abc ")
        'abc'
        >>> strip("def")
        'def'

    .. note::

        Because processors are type-aware they are most useful directly after type
        verification or coercion.
    """

    @abstractmethod
    def __call__(self, val: A) -> A:  # pragma: no cover
        raise NotImplementedError()  # pragma: no cover


@dataclass
class CacheValidatorBase(Validator[A]):
    """
    This class should be subclassed to work with whatever caching configuration
    is desired
    """

    validator: Validator[A]

    def cache_get_sync(self, val: Any) -> Maybe["ValidationResult[A]"]:
        """
        Try to get a value from a cache and return ``Maybe[ValidationResult[A]]``

        :param val: the value being validated
        """
        raise NotImplementedError()

    def cache_set_sync(self, val: Any, cache_val: "ValidationResult[A]") -> None:
        raise NotImplementedError()

    async def cache_get_async(self, val: Any) -> Maybe["ValidationResult[A]"]:
        """
        Try to get a value from a cache asynchronously and
        return ``Maybe[ValidationResult[A]]``

        :param val: the value being validated
        """
        raise NotImplementedError()

    async def cache_set_async(self, val: Any, cache_val: "ValidationResult[A]") -> None:
        raise NotImplementedError()

    def __call__(self, val: Any) -> "ValidationResult[A]":
        cache_result = self.cache_get_sync(val)
        if cache_result.is_just:
            return cache_result.val
        else:
            result = self.validator(val)
            self.cache_set_sync(val, result)
            return result

    async def validate_async(self, val: Any) -> "ValidationResult[A]":
        cache_result = await self.cache_get_async(val)
        if cache_result.is_just:
            return cache_result.val
        else:
            result = await self.validator.validate_async(val)
            await self.cache_set_async(val, result)
            return result



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Generic, Literal, Union

from koda_validate._generics import A, B
from koda_validate.errors import ErrType

if TYPE_CHECKING:
    from koda_validate.base import Validator


@dataclass
class Valid(Generic[A]):
    """
    A wrapper for valid data, e.g. ``Valid("abc")``
    """

    val: A
    """
    The value that has succeeded validation
    """

    is_valid: ClassVar[Literal[True]] = True
    """
    This is always ``True`` on :class:`Valid` instances. It's useful for ``if``
    statements. Mypy understands it as a tag for a tagged union.
    """

    def map(self, func: Callable[[A], B]) -> "ValidationResult[B]":
        return Valid(func(self.val))


@dataclass
class Invalid:
    """
    Represents validation failure. Contains relevant failure data so use case-specific
    error objects (or other data) can be produced.
    """

    err_type: "ErrType"
    """
    Any of a number of classes that contain data about the type of error, e.g.
    :class:`TypeErr`, :class:`CoercionErr`, :class:`KeyMissingErr`, etc.
    """

    value: Any
    """
    The invalid value that was being validated
    """

    validator: "Validator[Any]"
    """
    The validator that determined ``value`` to be invalid
    """

    is_valid: ClassVar[Literal[False]] = False
    """
    This is always ``False`` on :class:`Invalid` instances. It’s useful for ``if``
    statements. Mypy understands it as a tag for a tagged union.
    """

    def map(self, func: Callable[[Any], B]) -> "ValidationResult[B]":
        return self


ValidationResult = Union[Valid[A], Invalid]



from koda_validate._internal import _ToTupleStandardValidator


class BytesValidator(_ToTupleStandardValidator[bytes]):
    r"""
    Validate a value is a ``bytes``, and any extra refinement.

    If ``predicates_async`` is supplied, the ``__call__`` method should not be
    called -- only ``.validate_async`` should be used.

    Example:

    >>> from koda_validate import *
    >>> validator = BytesValidator(not_blank, MaxLength(100), preprocessors=[strip])
    >>> validator(b"")
    Invalid(err_type=PredicateErrs(predicates=[NotBlank()]), ...)
    >>> validator("")
    Invalid(err_type=TypeErr(expected_type=<class 'bytes'>), ...)
    >>> validator(b' ok ')
    Valid(val=b'ok')

    :param predicates: any number of ``Predicate[bytes]`` instances
    :param predicates_async: any number of ``PredicateAsync[bytes]`` instances
    :param preprocessors: any number of ``Processor[bytes]``, which will be run before
        :class:`Predicate`\s and :class:`PredicateAsync`\s are checked.
    :param coerce: a function that can control coercion
    """

    _TYPE = bytes
